
/********************************************************************************************
 * Home Ctrl is a professional IP based remote control system. Refer to
 * http://klaus30.bplaced.net/homectrl28/index.html This is the server application, 
 * it is controlled from an Android device. Requires Arduino Mega, EtherMega or 
 * compartible MC-board. Written by Klaus Pintoffl. Feel free to use, distribute 
 * share and modify this server program.
 ********************************************************************************************/

/********************************************************************************************
 * Function:  1  2  3  4  5  6  7  8  9 10 11 12 13 14 15 16 17 18 19 20 21 22 23 24 25 26 27 28 
 * Pin:      22 23 24 25 26 27 28 29 30 31 32 33 34 35 36 37 38 39 40 41 42 43 44 45 46 47 48 49
 ********************************************************************************************/

/********************************************************************************************
 * Function Properties: 
 * Char1            n=SaveState off        y=SaveState on 
 * Char2            n=none         c=confirm        r=random        p=pin 
 * Char3            n=no delay        s=4sec delay        l=8sec delay
 ********************************************************************************************/


#include <SPI.h>
#include <Ethernet.h>
#include <SD.h>
#include <EEPROM.h>
#include <TimerOne.h>
#include <avr/wdt.h>
#include <OneWire.h>
#include <Wire.h>
#include "RTClib.h"
#include <AESLib.h> 


//debug option used or not
boolean debug = true; 
//set false if Arduino with bootloader used (conflicts)
boolean watchdog = false;

//network parameters
byte IpArray[4]; 
byte mac[6];
byte GatewayArray[4];
byte SubnetArray[4];
int Port=0;

//outpins on Mega Board, see table on top
byte outPins[] = {9, 22, 23, 24, 25, 26, 27, 28, 29, 30, 31, 32, 33, 34, 35, 36, 37, 38, 39, 40, 41, 42, 43, 44, 45, 46, 47, 48, 49};
//delays in sec for auto-off. Index 0 is dummy
int Delay[29] = {0};
//substract from delays until 0. Index 0 is dummy  
int DelayCounter[29] = {0};
//seconds the alarm is on
int alarmDuration = 300;
//substract from alarmDuratian delays until 0 
int alarmDurationCounter = 300;
//the names of functions, like Light1...
char FNames[29][25];
//Array to describe SaveState, Security and AutoDisable for each function (incl dummy and terminator)
char FnProperties[30] = "daaaaaaaaaaaaaaaaaaaaaaaaaaaa";
//incoming string from client
String inString;
//4 Random bytes that have to be remembered between communications
byte randomArray[4];
//String for returning fn names and times/alarms
String FString;
//strong background key, and the user keys. All 16 bytes long
byte key[] = {107,99,89,10,79,109,105,102,116,112,80,113,86,106,77,116};
//the keys for all 16 users (rows) as 2D-Array. Last index 1 if Key read
byte UserKeys[16][17];
//byte aarray for message
byte m1[16];
//PINs can be given to each user
int PinArray[16] = {0};
//stores which user (row) can use which functions
byte UserFunctions[16][17] = {0}; //index 0 stores function #
//instance of oneWire on pin 2
OneWire ow(2);
//the currently used DS18S20 ID
byte addrC[8]; 
//max 5 DS18S20 devices possible
byte addr1[8]; 
//max 5 DS18S20 devices possible
byte addr2[8]; 
//max 5 DS18S20 devices possible
byte addr3[8]; 
//max 5 DS18S20 devices possible
byte addr4[8]; 
//results of sensor readings
byte DSdata[5][9];
//number of DS18S20 devices on OneWire bus
int DSNr = 0;  
//which DS is queried
int DSquery = 0;
//Real Time Clock object
RTC_DS1307 RTC1;
//DateTime object
DateTime now;                                            
//Number of output pins used
int FnNr = 28;
//the settings file version
byte Update = 0;
//the current user in UserFuncions[]
int row;
//if current user is an admin (A,B,C,D,E,F) this is true
boolean Admin = false;





void setup(){

  if (debug) {      
    Serial.begin(9600);
    Serial.println("Debug true");
  } 

  //if watchdog used, set it to 8 sec first
  if (watchdog) {
    MCUSR=0;
    wdt_enable(WDTO_8S);
  }

  // Initialize the SPI bus
  pinMode(53, OUTPUT); // set the SS pin as an output (necessary!) pin 53 on mega
  pinMode(10,OUTPUT); // set the w5100 SS pin to OUTPUT
  digitalWrite(10, HIGH); // temporarily turn off the W5100 chip! 

  //write defaults into function name arrays
  for (int x=1; x<=FnNr; x++) {
    String s = "Function " + (String)x;
    char xy[s.length()];     
    s.toCharArray(xy, s.length()+1);
    strcpy(FNames[x], xy);
  }


  //initialize I2C bus and RTC
  Wire.begin(); //used pin 20 and 21 on Mega
  RTC1.begin(); 
  //set clock to compile time. This is not needed as it can be done remotely
  //if (! RTC.isrunning()) RTC1.adjust(DateTime(__DATE__, __TIME__)); 

  //Temp sensors are searched on the bus. If present, up to 4 addresses are recorded
  while ((ow.search(addrC)) && (DSNr<5)) { //returns true if a device found   
    if ((addrC[0] == 0x10) && ( OneWire::crc8(addrC, 7) == addrC[7])) { //0x10=DS18S20/DS1820
      DSNr++; //Number of sensors
  
      for (int x=0; x<8; x++) {     
        switch (DSNr) {                   
          case 1:            
            addr1[x] = addrC[x];           
            break;                        
          case 2:     
            addr2[x] = addrC[x];
            break; 
          case 3:            
            addr3[x] = addrC[x];           
            break;                        
          case 4:     
            addr4[x] = addrC[x];
            break;                                
        }//end switch                               
      }//end for     
       
      } //end if 0x10 and CRC 
    } //end while
  

  //card has to be initialized for pin 4  
  if (!SD.begin(4)) {   
    if (debug) Serial.println("Card init failed, stopped");
    return;
  }

  //Read the network and user config from SD Card:
  File file1 = SD.open("Settings.txt");

  if (file1) {    
    char allText[file1.available()]; 
    int length = file1.available(); 

    //writes all bytes of text into the array allText
    for (int x=0; x<length; x++){      
      allText[x] = file1.read();           
    };

    char* parts;
    char* split = ",=.-:; \n\r";
    parts = strtok (allText, split);  

    while (parts != NULL){     
      String s = String(parts); 

      //if IP type, then the next 4 bytes are the IP adress
      if (s=="IP") { 
        for (int i= 0; i<4; i++){
          parts = strtok(NULL,split);        
          int param = atoi(parts); //atoi needs a char array
          IpArray[i] = param; 
        };        
      };

      //if MAC type, then the next 6 bytes are the MAC adress
      if (s=="MAC") { 
        for (int i= 0; i<6; i++){
          parts = strtok(NULL, split);
          int param = strtol (parts,NULL,16); //HEX base!                                                                                                           
          mac[i] = param;             
        };         
      };

      //if Gateway type, then the next 4 bytes are the Gateway adress
      if (s=="Gateway") {
        for (int i= 0; i<4; i++){
          parts = strtok(NULL, split); 
          int param = atoi(parts);                                                       
          GatewayArray[i] = param;         
        }                 
      };    

      //if Subnet type, then the next 4 bytes are the Subnet adress
      if (s=="Subnet") {
        for (int i= 0; i<4; i++){
          parts = strtok(NULL, split);
          int param = atoi(parts);                                                     
          SubnetArray[i] = param;         
        }        
      };  

      //if Port type, then the next byte is the Port
      if (s=="Port") { 
        parts = strtok(NULL, split); 
        Port = atoi(parts);  
      };

      //for the Delays 1 to 28:
      if(s.indexOf("Delay")>-1) {
        for (int x=1; x<=FnNr; x++){
          if (s=="Delay"+String(x)) {
            parts = strtok(NULL, split);
            Delay[x] = DelayCounter[x] = atoi(parts);
            break; 
          }        
        }       
      }


      //if Key, one round reads this code
      if(s.startsWith("Key")) {         
        char userChar = s[3];                  
        parts = strtok(NULL, split);
        String s = String(parts);//only for getting length
        int loops = s.length();        
                                                                               
        if (loops>16)loops = 16; //max 16 digits of user key used, rest ignored    
        row = getUserRow(userChar); //function call userchar in row
                
        for (int x=0; x<loops; x++) {
          UserKeys[row][x] = (byte)(parts[x]);          
        }; 
     
        for (int x=loops; x<16; x++) { 
          UserKeys[row][x] = (byte)(key[x]);
        };
        
        UserKeys [row][16] = 1; // 1 means the user is allowed
      };


      //Functions available for each user, and their number 
      if(s.startsWith("Functions")) {         
        char userChar = s[9];  
        row = getUserRow(userChar); //function call    
        char* parts2; 
        parts2 = strtok(NULL,"=.\n"); //do not stop at "," or " "      
        String s = String(parts2); //s now contains e.g "1,3,4,5,7"        
        int index, fnCount = 0;       
        while (s.length()>0) {
          fnCount++; //gets 1 for the 1st function
          if (fnCount>16) break; //only 16 functions allowed per user, no crash
          if (s.indexOf(",")>0) index = s.indexOf(",");
          else index = s.length();    
          String s2 = s.substring (0, index); //contains the current function number
          s = s.substring(index+1 ,s.length()); //cut s by s2                    
          char xy[s2.length()]; 
          s2.toCharArray(xy, s2.length()+1);
          UserFunctions[row][fnCount] = atoi(xy); //write function into array      
        } 
        if (fnCount>16) fnCount=16; //reduce by one if it was 17! 
        UserFunctions[row][0] = fnCount; //index 0 is number of functions for userChar      
      };


      //for the Function Names 
      if (s.startsWith("Name")) {                
        for (int x=1; x<=FnNr; x++){         
          if (s=="Name"+String(x)) {
            parts = strtok(NULL, "\n\r"); 
            String s = String(parts); //for the length          
            if (s.length()<25) strcpy(FNames[x], parts); 
            break; 
          } 
        }  
      }   


      //if s is Properties. A code for each property combination is used 
      if (s.startsWith("Properties")) {                
        for (int x=1; x<=FnNr; x++){         
          if (s=="Properties"+String(x)) {
            parts = strtok(NULL, split);
            String s = String(parts);                    
            if (s=="nnn") FnProperties[x] = 'a';  if (s=="ynn") FnProperties[x] = 'm';
            if (s=="nns") FnProperties[x] = 'e';  if (s=="yns") FnProperties[x] = 'q';
            if (s=="nnl") FnProperties[x] = 'i';  if (s=="ynl") FnProperties[x] = 'u';
            if (s=="ncn") FnProperties[x] = 'b';  if (s=="ycn") FnProperties[x] = 'n';
            if (s=="ncs") FnProperties[x] = 'f';  if (s=="ycs") FnProperties[x] = 'r';
            if (s=="ncl") FnProperties[x] = 'j';  if (s=="ycl") FnProperties[x] = 'v';
            if (s=="nrn") FnProperties[x] = 'c';  if (s=="yrn") FnProperties[x] = 'o';
            if (s=="nrs") FnProperties[x] = 'g';  if (s=="yrs") FnProperties[x] = 's';
            if (s=="nrl") FnProperties[x] = 'k';  if (s=="yrl") FnProperties[x] = 'w';
            if (s=="npn") FnProperties[x] = 'd';  if (s=="ypn") FnProperties[x] = 'p';
            if (s=="nps") FnProperties[x] = 'h';  if (s=="yps") FnProperties[x] = 't';
            if (s=="npl") FnProperties[x] = 'l';  if (s=="ypl") FnProperties[x] = 'x';                                                                   
            break; 
          } 
        }                                                                                       
      }
      
      
   

      //if s is PIN 
      if (s.startsWith("PIN")) { 
        char userChar = s[3];
        row = getUserRow(userChar); //function call          
        parts = strtok(NULL, split);           
        PinArray[row] = atoi(parts);                                                                                                                                                                
      } 



      //for Update (Version of Settings file) 
      if (s=="Update") {                
        parts = strtok(NULL, "\n\r"); 
        Update = atoi(parts);                        
      }


      parts = strtok(NULL, split);   
    }//end while (parts != NULL)

    file1.close(); 
    //Network config read from SD Card done.
  }  

  //SaveState wins over Delay: Both at the same time does not make sense 
  for (int x=1; x<=FnNr; x++){    
    if ((FnProperties[x]>=109) && (FnProperties[x]<=120)) { //means char 'm' to 'x', with saveState                                                           
      Delay[x]=0;
      DelayCounter[x]=0;          
    }
  }

  //set the pins 22 to 49 as outputs. 0 is ISR indicator
  for (int x=0; x<=FnNr; x++){                           
    pinMode(outPins[x], OUTPUT);                                                                                       
  }

  //set Alarmpin (6) and Timepins (7,8) as output
  pinMode(6, OUTPUT); //alarm
  pinMode(7, OUTPUT); //timepair 1
  pinMode(8, OUTPUT); //timepair 2

  //Check SaveStates and EEProms and switch outputs on if needed:
  for (int x=1; x<=FnNr; x++) {     
    if ((FnProperties[x]>=109) && (FnProperties[x]<=120) && (EEPROM.read(x)==1)) digitalWrite (outPins[x], HIGH);                   
  }

  //get alarms and times from EEPROM into timeArray. Check against 255 (addresses not used yet)
  for (int x=100; x<124; x++) {
    if (EEPROM.read(x)>160) {
      for (int y=100; y<124; y++) {
        EEPROM.write(y, 0);
      }
      break; 
    }      
  }    

  //Uses TimerOne Library. Every second ISR1 is called.
  Timer1.initialize(1000000); 
  Timer1.attachInterrupt(ISR1);

  Ethernet.begin(mac, IpArray,GatewayArray,SubnetArray);

  //avoids mem fragmentation when appending to Strings
  inString.reserve(70);
  FString.reserve(500);

}//end setup










void loop() {

  //reset watchdog
  if (watchdog) wdt_reset();

  //has both to be done here, after Port read  
  EthernetServer server(Port); 
  server.begin();  

  EthernetClient client = server.available();

  if (client) {

    //http request ends with a blank line
    boolean current_line_is_blank = true; 
    while (client.connected()) {

      if (client.available()) {
        char c = client.read();
        //50 is 8 byte more than actually needed
        if (inString.length() < 50) { 
          inString.concat(c);
        } 

        if (c == '\n' && current_line_is_blank) {
          //Serial.println(inString);
          //send standard http response header
          client.println("HTTP/1.1 200 OK");
          client.println("");
          client.println();          

          if(inString.indexOf("*")>0){ 
            
            inString = inString.substring(inString.indexOf("*")+1, inString.lastIndexOf("*"));                                          
                                                        
            if (inString.length()%32 == 1) {   
              char userChar = inString[0];           
              row = getUserRow (userChar);
              
              //write correct key in key[]
              for (int x=0; x<16; x++) {
                key[x] = UserKeys[row][x];          
              };      
                      
              inString = inString.substring(1, inString.length());           
            } 

          if (UserKeys[row][16] == 1) { //only if user has a key
                                                          
            //Call to convert HEX string into byte array m1
            HexStringIntoByteArray (inString); 
            //Call to decode the byteArray (built in function)              
            aes128_dec_single(key, m1);


            //if Status is obtained by client***************************************************************
            if ((char)m1[0]==('S')) { 
         
              writeNewRandomArray(); //refresh the random array           
              m1[9] = UserFunctions[row][0]; //Write the function # for user row                                                                             
              m1[13] = Update;           
              PinStatusIntoArray(); //get Status of pins and write as bytes 10,11 into array          
              aes128_enc_single(key, m1); //encode the byteArray              
              String coded1 = returnHexFromByteArray (m1); //Hex string generation              
              client.print(coded1); //send to client       
            };//end if Status Request


            //if only the ID is requested**********************************************************************
            if ((char)m1[0]==('I')) { 
      
              writeNewRandomArray(); //refresh the random array
              aes128_enc_single(key, m1);
              String coded1 = returnHexFromByteArray (m1); 
              client.print(coded1);             
            }



            //if function execution requested*******************************************************************
            if ((char)m1[0]==('E')) { 

              if (checkOk()){ //execute the function if checkOk returns true

                m1[0] = 'S'; //Status     
                int toExe = m1[12]; //which function to execute
                m1[13] = Update;  
                int fnToUse = UserFunctions[row][toExe]; //the actual function to switch
                
                //if PIN matches, set the output pin, and EEPROM if SaveState true
                if (pinOk(fnToUse)) {                
                     
                    if (digitalRead(outPins[fnToUse])) { //if the pin is high               
                      digitalWrite(outPins[fnToUse],LOW);
                      if ( (FnProperties[fnToUse]>=109) && (FnProperties[fnToUse]<=120) ) EEPROM.write(fnToUse , 0);                                   
                    } 
                    
                    else { //if the pin is low       
                      digitalWrite(outPins[fnToUse],HIGH);
                      if ( (FnProperties[fnToUse]>=109) && (FnProperties[fnToUse]<=120) ) EEPROM.write(fnToUse , 1);
                    }
                    
                    m1[14] = m1[15] = 0; //replace a correct pin with 0
                } //end pinCheck(int,int)                 
    
                m1[9] = UserFunctions[row][0]; //Write the function # for user row          
                PinStatusIntoArray(); //get Status of pins and write as bytes into array     
                aes128_enc_single(key, m1); //encode the byteArray     
                String coded1 = returnHexFromByteArray (m1); //Hex string generation  
                client.print(coded1);                 
              } //end checkOk()           
            }//if function execution requested


            //if RTC Set requested by client (Admin only)******************************************************
            if ((char)m1[0]==('T')) { 
              //time info on indices 9 to 14: hh,mm,ss,dd,MM,yy

              if (checkOk()) { //set time if checkOk returns true

                if ((m1[15] == 0) && (Admin)) {//if RCT set required, and User is Admin
                  int timeValue;
                  byte set;        
                  Wire.beginTransmission(0x68); //start the bus tranmission to set time            
                  Wire.write(0);
                  
                  timeValue = m1[11];  
                  set = timeValue/10*16 + timeValue%10;  
                  Wire.write (set); //seconds         
                  timeValue = m1[10];  
                  set = timeValue/10*16 + timeValue%10;  
                  Wire.write (set); //minutes     
                  timeValue = m1[9];   
                  set = timeValue/10*16 + timeValue%10;  
                  Wire.write (set); //hours     
                  timeValue = 1;       
                  set = timeValue/10*16 + timeValue%10;  
                  Wire.write (set); //why?                           
                  timeValue = m1[12];  
                  set = timeValue/10*16 + timeValue%10;  
                  Wire.write (set); //day       
                  timeValue = m1[13];  
                  set = timeValue/10*16 + timeValue%10;  
                  Wire.write (set); //month   
                  timeValue = m1[14];  
                  set = timeValue/10*16 + timeValue%10;  
                  Wire.write (set); //year, only 2 digits  
                  
                  Wire.write(0x10); // sends 0x10 (hex) 00010000 (binary) to control register - turns on square wave
                  Wire.endTransmission(); 
                }

                if (m1[15] == 1) {//Alarm or time to set to AlarmTime Array and EEPROM
                  int whichAlarm = m1[11];
                  int alRow;

                  switch (whichAlarm) {
                  case 0: 
                    alRow = 0; 
                    break;
                  case 1: 
                    alRow = 1; 
                    break;
                  case 2: 
                    alRow = 2; 
                    break;
                  case 3: 
                    alRow = 3; 
                    break;
                  case 100: 
                    alRow = 4; 
                    break;
                  case 101: 
                    alRow = 5; 
                    break;
                  case 102: 
                    alRow = 6; 
                    break;
                  case 103: 
                    alRow = 7; 
                    break;                  
                  }
                
                  if (m1[12] == 1) {//set new time
                    EEPROM.write(100+alRow*3, 1);
                    EEPROM.write(101+alRow*3, m1[9]);
                    EEPROM.write(102+alRow*3, m1[10]);      
                  }


                  if (m1[12] == 2) {//keep old time, enable
                    EEPROM.write(100+alRow*3, 1);
                  }

                  if (m1[12] == 0) {//keep old time, disable
                    EEPROM.write(100+alRow*3, 0);
                  }     
                }

                byte ClientChecksum [4]; 
                //M1 gets overwritten on each loop!

                for (int x=0; x<4; x++) {
                  ClientChecksum [x] = m1[x+1];   
                }
          
                FString = ""; //globaly defined
                int hourCode; //adds 100 to hour value if alarm is on
                
                //a time object is generated every second in the interrupt service routine
                FString.concat(now.hour() + (String)"*" + now.minute() + (String)"*" + now.second() + (String)"#"); //Server Time
                FString.concat(now.dayOfWeek() + (String)"*" + now.month() + (String)"*" + now.day() + (String)"*" + now.year() + (String)"#"); //Server Date

                for (int x=0; x<8; x++) {
                  if (EEPROM.read(100 + x*3)==1) hourCode = EEPROM.read(101 + x*3) + 100; 
                  else hourCode = EEPROM.read(101 + x*3);   
                  FString.concat(hourCode + (String)"*" + EEPROM.read(102 + x*3));
                  FString.concat("#"); //Separator char                                                    
                }
                                                                                                         
                int len = FString.length(), y=0;                                                                                          
                char xy[FString.length()]; 
                FString.toCharArray(xy, len+1); 

                while (y<len) { //loop to write String to m1 indices 6-15      
                  m1[0] = 'T';

                  for (int x=0; x<4; x++) {
                    m1[x+1] = ClientChecksum [x];   
                  }

                  for (int x=5; x<16; x++) {
                    if (y<len) m1[x] = xy[y]; 
                    else break;
                    y++;
                  }

                  aes128_enc_single(key, m1); //encode the byteArray              
                  String coded1 = returnHexFromByteArray (m1); //Hex string generation              
                  client.print(coded1); //send to client                                      
                }//end while                              
              }//checkOk                     
            }


            //if Function name update requested**********************************************************************
            if ((char)m1[0]==('F')) { 
    
              if (checkOk()) {
                byte ClientChecksum [4]; 

                for (int x=0; x<4; x++) {
                  ClientChecksum [x] = m1[x+1];   
                }

                FString = "";

                for (int x=1; x<=UserFunctions[row][0]; x++) {
                  FString.concat(FnProperties[UserFunctions[row][x]]);// add property code
                  FString.concat(FNames[UserFunctions[row][x]]); //add the Fn name
                  FString.concat("#"); //Separator char                                                    
                }
                                                        
                int len = FString.length(), y=0;                                    
                char xy[FString.length()]; 
                FString.toCharArray(xy, len+1); 

                while (y<len) { //loop to write String to m1 indices 6-15       
                  m1[0] = 'F';

                  for (int x=0; x<4; x++) {
                    m1[x+1] = ClientChecksum [x];   
                  }

                  m1[5] = UserFunctions[row][0]; //fn# on index 5 (not 9 as usual) 

                  for (int x=6; x<16; x++) {
                    if (y<len) m1[x] = xy[y]; 
                    else break;
                    y++;
                  }
                  Serial.println((char)m1[0]);
                  aes128_enc_single(key, m1); //encode the byteArray              
                  String coded1 = returnHexFromByteArray (m1); //Hex string generation              
                  client.print(coded1); //send to client                                 
                  Serial.println(coded1);   
                }//end while
              }//end ich checkOk              
            } 
                      
            
            //if Values (V) or Value names sync (v) requested**********************************************************************
            if ((char)m1[0]==('V') || (char)m1[0]==('v')) { 
   
              if (checkOk()) {     
                char valueMode = m1[0]; //Values or Value names
                byte ClientChecksum [4]; 

                for (int x=0; x<4; x++) {
                  ClientChecksum [x] = m1[x+1];   
                }

                FString = ""; //global var
                String ValueName; 
                ValueName.reserve(60);
                String Value;
                Value.reserve(60);
                         
                for (int x=1; x<=12; x++) {               
                  ValueName = "";
                  Value = "";
                  
                  switch (x) {
                    
                    // User Values ***** User Values ****** User Values ***** User Values ***** User Values *****  
                    case 1: 
                    ValueName = "Server Time";
                    Value = (String) now.hour() + ":" + (String) now.minute() + ":" + (String) now.second();   
                    break;
                    
                    case 2: 
                    ValueName = "Analog reading pin 3";
                    Value = (String) analogRead(3);
                    break;
                    
                    case 3: 
                    ValueName = "Analog Read pin 4";
                    Value = (String) analogRead(4);
                    break;
                    
                    case 4: 
                    //ValueName = "Another value";
                    //Value = "Test Value";
                    break;
                    
                    case 5: 
                    if (Admin) { //only Admins will see that value
                      float Volts = (analogRead(5)*5.00/1023.00);       
                      char tmp[8];  
                      ValueName = "Server Vcc [Volts]"; //dtostrf (val, width, prec, s) 		
                      Value = dtostrf(Volts, 5, 2, tmp);      
                    }
                    break;                
                    
                    case 6: 
                    ValueName = "Temp DS18S20 1 [Deg C]";
                    Value = getDS18S20 (1);            
                    break;
                                      
                    case 7: 
                    ValueName = "Temp DS18S20 2 [Deg C]";
                    Value = getDS18S20 (2);            
                    break;
                    
                    case 8: 
                    //ValueName = "Temp DS18S20 3 [Deg C]";
                    //Value = getDS18S20 (3);            
                    break;
                    
                    
                    case 9: 
                    //ValueName = "Temp DS18S20 4 [Deg C]";
                    //Value = getDS18S20 (4);            
                    break;
                    
                    case 10: break;
                    case 11: break;
                    case 12: break; //use as many as you want  
                    // User Values ***** User Values ****** User Values ***** User Values ***** User Values ***** 
                  }
                  
                  if (ValueName != "") {
                    if (valueMode == 'V') {
                      FString.concat(Value + "#");}  
                    else {   
                      FString.concat(ValueName + "#");
                      FString.concat(Value + "#");
                    }
                  }   
                }//end for loop user values
                                                                        
                int len = FString.length(), y=0;                                    
                char xy[FString.length()+1]; 
                FString.toCharArray(xy, len+1); 

                while (y<len) { //loop to write String to m1 indices        
                  m1[0] = valueMode;

                  for (int x=0; x<4; x++) {
                    m1[x+1] = ClientChecksum [x];   
                  }

                  for (int x=5; x<16; x++) {
                    if (y<len) m1[x] = xy[y]; 
                    else break;
                    y++;
                  }
                
                  aes128_enc_single(key, m1); //encode the byteArray              
                  String coded1 = returnHexFromByteArray (m1); //Hex string generation              
                  client.print(coded1); //send to client                                 
                                                                                 //Serial.println(coded1);   
                }//end while
              }//end ich checkOk              
            } 
            
                  
            //if Pro-Tab values requested**********************************************************************
            if (((char)m1[0]==('P')) && (Admin)) { 
              
              if (checkOk()) {      
                writeNewRandomArray(); //refresh the random array
                
                for(int x=0; x<7; x++) {
                  m1[x+9] = EEPROM.read(x+200);
                }
                      
                aes128_enc_single(key, m1);
                String coded1 = returnHexFromByteArray (m1); 
                client.print(coded1); 
              }            
            }
            
            
            
            //if Pro-Tab values received**********************************************************************
            if ((char)m1[0]==('Q')) {
             
             if (checkOk()) {  
                writeNewRandomArray(); //refresh the random array       
                m1[0]='P';
                
                for (int x=0; x<7; x++) {
                  if (EEPROM.read(x+200) != m1[x+9]) 
                    EEPROM.write(x+200, m1[x+9]);         
                }
                           
                aes128_enc_single(key, m1);
                String coded1 = returnHexFromByteArray (m1); 
                client.print(coded1);
             }             
            }
                        
            
            //if kill alarm on pin 6 requested by client******************************************************
            if ((char)m1[0]==('K')) { 
              
              if (checkOk()) {                                                                                
                digitalWrite (6, LOW);
                alarmDurationCounter = 0;
                  
                aes128_enc_single(key, m1); //encode the byteArray              
                String coded1 = returnHexFromByteArray (m1);              
                client.print(coded1); 
              }       
            };
          } //end if UserKeys[row][16] == 1                         
          }//end if(inString.indexOf("*")>0)
          break;
        }

        //starting a new line:
        if (c == '\n') {
          current_line_is_blank = true;
        } 
        //a character on the current line
        else if (c != '\r') {         
          current_line_is_blank = false;
        }
      }
    }
   
    delay(1); //gives client time to receive the data (needed?) 
    inString = ""; //reset the received string
    client.stop();
  }  
}//end loop 






//interrupt service routine to handle delays. Called every second.
void ISR1(){                                                                         

  // For the Temperature:   
  ow.reset(); //reset bus with all devices
   
  if ((DSNr>0) && (DSquery <= DSNr)) {     
    switch (DSquery) {    
      case 1: 
        ow.select(addr1);
        ow.write(0xBE); //read the scratchpad      
        for (byte i=0; i<9; i++) {DSdata[1][i] = ow.read();} 
        ow.reset(); 
        ow.select(addr1);
        ow.write(0x44,1); //write data to scratchpad
        break;     
      case 2: 
        ow.select(addr2);
        ow.write(0xBE); //read the scratchpad      
        for (byte i=0; i<9; i++) {DSdata[2][i] = ow.read();} 
        ow.reset(); 
        ow.select(addr2);
        ow.write(0x44,1); //write data to scratchpad
        break;    
      case 3: 
        ow.select(addr3);
        ow.write(0xBE); //read the scratchpad      
        for (byte i=0; i<9; i++) {DSdata[3][i] = ow.read();} 
        ow.reset(); 
        ow.select(addr3);
        ow.write(0x44,1); //write data to scratchpad
        break;    
      case 4: 
        ow.select(addr4);
        ow.write(0xBE); //read the scratchpad      
        for (byte i=0; i<9; i++) {DSdata[4][i] = ow.read();} 
        ow.reset(); 
        ow.select(addr4);
        ow.write(0x44,1); //write data to scratchpad
        break;      
    } //end switch         
    ow.reset(); //reset bus with all devices
  } //end if  
  
  DSquery++; 
  if (DSquery > 4) DSquery = 0; //every sensor read all 5 sec if 4 connected
  //////////////////////////////////////////////////////////////////

  //toggle control led 
  if (digitalRead(outPins[0])) digitalWrite(outPins[0], LOW);  
  else digitalWrite(outPins[0], HIGH);

  //Check and decrement function delays
  for (int x=1; x<=FnNr; x++){ 
    if (Delay[x] >0) {
      if ((DelayCounter[x]>0) && (digitalRead(outPins[x])==HIGH)){ 
        DelayCounter[x]--;
      } 
      else { 
        digitalWrite(outPins[x], LOW);
        DelayCounter[x]=Delay[x];
      }
    }
  }
 
  //check alarms and times, swith pins if needed
  now = RTC1.now();
  long onTime; //exceeds range of int
  long offTime; 
  long nowTime = (now.hour()*60L + now.minute())*60L + now.second();

   //daily alarm
   if ((EEPROM.read(100)) && (EEPROM.read(101)==now.hour()) && (EEPROM.read(102)==now.minute()) && (now.second()<5)){      
     alarmDurationCounter = alarmDuration;
     digitalWrite (6, HIGH);        
   }
      
   //weekday alarm
   if ((EEPROM.read(103)) && (EEPROM.read(104)==now.hour()) && (EEPROM.read(105)==now.minute()) && (now.second()<5) && (now.dayOfWeek()>0) && (now.dayOfWeek()<6)){      
     alarmDurationCounter = alarmDuration;
     digitalWrite (6, HIGH);        
   }
     
   //weekend alarm
   if ((EEPROM.read(106)) && (EEPROM.read(107)==now.hour()) && (EEPROM.read(108)==now.minute()) && (now.second()<5) && (now.dayOfWeek()==0 || now.dayOfWeek()==6)){      
     alarmDurationCounter = alarmDuration;
     digitalWrite (6, HIGH);        
   }

   //single alarm
   if ((EEPROM.read(109)) && (EEPROM.read(110)==now.hour()) && (EEPROM.read(111)==now.minute()) && (now.second()<5)){      
     alarmDurationCounter = alarmDuration;
     digitalWrite (6, HIGH);
     EEPROM.write(109 , 0);        
   }
              
  //time pair 1    
  if (EEPROM.read(112) && EEPROM.read(115)) { 
    onTime = (EEPROM.read(113)*60L + EEPROM.read(114))*60L;  
    offTime = (EEPROM.read(116)*60L + EEPROM.read(117))*60L; 
    
    if (onTime<offTime) {    
      if ((onTime<nowTime) && (nowTime<offTime)) digitalWrite (7, HIGH); else digitalWrite (7, LOW);
    }
              
    if (onTime>offTime) {
      if ((offTime<nowTime) && (nowTime<onTime)) digitalWrite (7, LOW); else digitalWrite (7, HIGH);
    }
  }
  else digitalWrite (7, LOW);
      
  // time pair 2    
  if (EEPROM.read(118) && EEPROM.read(121)) { 
    onTime = (EEPROM.read(119)*60L + EEPROM.read(120))*60L;                             
    offTime = (EEPROM.read(122)*60L + EEPROM.read(123))*60L;   
     
    if (onTime<offTime) {
      if ((onTime<nowTime) && (nowTime<offTime)) digitalWrite (8, HIGH); else digitalWrite (8, LOW);
    }
        
    if (onTime>offTime) {
      if ((offTime<nowTime) && (nowTime<onTime)) digitalWrite (8, LOW); else digitalWrite (8, HIGH);
    }
  } 
  else digitalWrite (8, LOW);    
  
  alarmDurationCounter--; 

  if (alarmDurationCounter<=0) {
    alarmDurationCounter = alarmDuration; //reset counter
    digitalWrite (6, LOW); 
  }                                                                     
}//end ISR1






// writes new random numbers for chechsum to m1
void writeNewRandomArray() {
  randomSeed(millis());

  for (int x=0; x<4; x++){    
    randomArray[x] = m1[x+5] = random(255);
  }   
 
}



//function returns a hexadecimal string of the given byte array. No negative values are calculated,as 128 is added to each char value
String returnHexFromByteArray (byte toConvertToHex []) {  
  String hexInterpret; 
  hexInterpret.reserve(100);  
  hexInterpret = "";   
  for (int x=0; x<16; x++){             
    //int a = toConvertToHex[x]+128;
    int a = toConvertToHex[x];
    String b = String (a, HEX);
    if(b.length()==1) b="0"+b;               
    hexInterpret.concat(b);                                                
  }    
  return hexInterpret;
}



//takes a String and writes it into byteArray m1
void HexStringIntoByteArray ( String in) {
  for (int x=0; x<32; x+=2) {    
    char char1 = in[x];                     
    char char2 = in[x+1];
    char hexDigit1 = toupper(char1);
    char hexDigit2 = toupper(char2);
    int digit1 = (hexDigit1 >= 'A') ? hexDigit1 - 'A' + 10 : hexDigit1 - '0';
    int digit2 = (hexDigit2 >= 'A') ? hexDigit2 - 'A' + 10 : hexDigit2 - '0';
    int final = digit1*16 + digit2;                                            
    m1[x/2] = final; //write it into byte array
  } 
}



// returns the on-off status of the output pins as bytes
void PinStatusIntoArray() {
  byte Status1 = 0, Status2 = 0;              
  int end = UserFunctions[row][0]; //the # of functions for user row

  for (int x=0; x<end; x++) { 
    if (x<8) bitWrite (Status1,  x,  digitalRead (outPins[UserFunctions[row][x+1]]));  
    else bitWrite (Status2,  x-8,  digitalRead (outPins[UserFunctions[row][x+1]]));
  }
  //write the staus into the array
  m1[10] = Status1; 
  m1[11] = Status2; 
}



// returns the row for the current user, and sets the global flag Admin
int getUserRow (char userChar) {   
  Admin = true;
   
  switch (userChar) {
    case 'A': 
      row = 10; 
      break;
    case 'B': 
      row = 11; 
      break;
    case 'C': 
      row = 12; 
      break; 
    case 'D': 
      row = 13; 
      break;
    case 'E': 
      row = 14; 
      break;
    case 'F': 
      row = 15; 
      break;      
    default: //row is the same as userChar in int 
      char xy[1];
      xy[0] = userChar;
      row = atoi(xy);
      Admin = false;    
  }  
  return row;          
}




boolean checkOk () {
  boolean check = true;   
  for (int x=0; x<4; x++) {
    if (m1[x+5] != randomArray[x]) {
      check = false; 
      if(debug) Serial.println("Checksum failed!"); 
      break;
    }
  }  
  return check;    
}



boolean pinOk (int Fn) { 
  boolean check = true;    
    if ( (FnProperties[Fn] == 'd') || (FnProperties[Fn] == 'h') || (FnProperties[Fn] == 'l') || (FnProperties[Fn] == 'p') || (FnProperties[Fn] == 't') || (FnProperties[Fn] == 'x')) {                                   
        if (m1[14]*256 + m1[15] != PinArray[row]) {
          check = false;   
        }
        if (PinArray[row]==0) check = false; //Pin may not be 0 (unchanged from init.)        
    }   
  return check;    
}



//returns the temperature as String 
String getDS18S20 ( int DSIndex) { 
  int tr;       
  tr = DSdata[DSIndex][0]; 
  //check for negative temperature
  if (DSdata[DSIndex][1] > 0x80) {
    tr = !tr + 1; //two's complement adjustment
    tr = tr * -1; //flip value negative.
  }
  int cpc = DSdata[DSIndex][7]; //COUNT PER Celsius degree (10h) 
  int cr = DSdata[DSIndex][6]; //COUNT REMAIN (0Ch)
  tr = tr >> 1; //drop bit 0
  float temp = tr - (float)0.25 + (cpc - cr)/(float)cpc;                 
  String a;
  char tmp[8];  
  dtostrf(temp, 7, 1, tmp);  
  a =  tmp;  
  return a;  
}


